from pattmatch.helpers import longest_proper_prefixes


def kmp(text, pattern): # +DOCTEST ELIPSIS
    """
    Finds all start-end of searched pattern in the given text.

    >>> kmp('ABACD', 'ACD')
    [[2, 5]]
    >>> kmp('ABCDABC', 'ABC')
    [[0, 3], [4, 7]]
    >>> kmp('AAAAABAAABA', 'AAAA')
    [[0, 4], [1, 5]]
    >>> kmp('ABC', 'D')
    []

    Args:
        text(str) : text for searching in
        
        pattern(str): pattern for searching in the text

    Returns all start-end positions of the pattern in the given text.
    """
    lps = longest_proper_prefixes(pattern)
    start_end_positions = []

    i, j = [0] * 2

    while i < len(text):
        if text[i] == pattern[j]:
            i = i + 1
            j = j + 1
        else:
            if j is 0:
                i = i + 1
            else:
                j = lps[j - 1]

        if j == len(pattern):
            start_end_positions.append([i - j, i])
            j = lps[j - 1]

    return start_end_positions




"""
Helper methods used from implemented algorithms.
"""

def difference(text, pattern, index): # +DOCTEST ELIPSIS
    """
    Method used by boyer moore's algorithm.
    """
    j = index - 1
    from_index = 0
    for i in range(len(pattern) - 1, -1, -1):
        if pattern[i] != text[j]:
            if not text[j] in pattern:
                return len(pattern) - from_index
            else:
                return i - pattern[:i].rindex(text[j])
            break
        from_index = from_index + 1
        j = j - 1
    return 0


def longest_proper_prefixes(pattern):
    """
    Returns longest common prefices of the given pattern.
    Method used by kmp's algorithm.

    >>> longest_proper_prefixes('ABCDABD')
    [0, 0, 0, 0, 1, 2, 0]
    >>> longest_proper_prefixes('AABAACAABAA')
    [0, 1, 0, 1, 2, 0, 1, 2, 3, 4, 5]
    >>> longest_proper_prefixes('AAAA')
    [0, 1, 2, 3]
    >>> longest_proper_prefixes('AAACAAAAAC')
    [0, 1, 2, 0, 1, 2, 3, 3, 3, 4]
    """
    longest_proper_prefix = [0] * (len(pattern))
    for current_index in range(len(pattern) + 1):
        longest_prefix = [pattern[:current_index - i] for i in range(1, current_index)
                           if pattern[:current_index - i] == pattern[i:current_index]]
        longest_proper_prefix[current_index - 1] = len(longest_prefix[0]) if len(longest_prefix) else 0

    return longest_proper_prefix

CHAR_SIZE = 256

def init_hash(i, j, text, prime):
    """
    Returns a hash of the text[i:j].
    Method used by rabin karp's algorithm.
    """
    global CHAR_SIZE
    text_hash = 0

    for l in range(i, j):
        text_hash = (CHAR_SIZE * text_hash + ord(text[l])) % prime
    return text_hash
 
def random_prime():
    """
    Returns random prime.
    Method used by rabin karp's algorithm.
    """
    random.seed()
    global CHAR_SIZE

    prime = None
    start_from = random.randint(CHAR_SIZE ** 2, CHAR_SIZE ** 3)
    for i in range(start_from, CHAR_SIZE ** 3):
        is_not_prime = sum([j for j in range(2, math.ceil(math.sqrt(i))) if i % j == 0])
        if is_not_prime:
            next
        else:
            prime = i
            break

    return prime


def longest_prefix(text, pattern, start):
    """
    Finds the longest prefix of text and pattern.
    """
    end = len(pattern)
    while text[start:start + end] != pattern[:end]:
        end -= 1

    return text[start:start + end]



import random
import math

from pattmatch.helpers import random_prime, init_hash

CHAR_SIZE = 256

def rabin_karp(text, pattern): # +DOCTEST ELIPSIS
    """
    Finds all start-end position of searched pattern in the given text.

    >>> rabin_karp('ABCD', 'ABC')
    [[0, 3]]
    >>> rabin_karp('ABC', 'ABC')
    [[0, 3]]
    >>> rabin_karp('ABABC', 'ABC')
    [[2, 5]]
    >>> rabin_karp('ABDBABCDCABC', 'ABC')
    [[4, 7], [9, 12]]

    Args:
        text(str) : text for searching in

        pattern(str): pattern for searching in the text

    Returns all start-end positions of the pattern in the given text.
    """
    global CHAR_SIZE

    prime = random_prime()
    power = CHAR_SIZE ** (len(pattern) - 1)

    text_hash = init_hash(0, len(pattern), text, prime)
    patt_hash = init_hash(0, len(pattern), pattern, prime)

    if text_hash == patt_hash and len(text) == len(pattern):
        return [[0, len(text)]]

    start_end_positions = []

    n, m = len(text), len(pattern)

    for i in range(n - m + 1):
        if text_hash == patt_hash and text[i:i+m] == pattern:
            start_end_positions.append([i, i+m])
        if i < n - m:
            text_hash = (CHAR_SIZE * (text_hash - ord(text[i]) * power) + ord(text[i + m])) % prime
    
    return start_end_positions



from pattmatch.helpers import longest_prefix


def z_array(text, pattern): # +DOCTEST ELIPSIS
    """
    Finds all occurences of a pattern in a text.

    >>> z_array('AAAB', 'AA')
    [2, 2, 1, 0]
    >>> z_array('AAABAAAAACAAAAAABAAAAAABBBAAA', 'AAAAB')
    [3, 2, 1, 0, 4, 4, 3, 2, 1, 0, 4, 4, 5, 3, 2, 1, 0, 4, 4, 5, 3, 2, 1, 0, 0, 0, 3, 2, 1]
    >>> z_array('ABCDAABBCAAAABBCAAAAABCA', 'AAAA')
    [1, 0, 0, 0, 2, 1, 0, 0, 0, 4, 3, 2, 1, 0, 0, 0, 4, 4, 3, 2, 1, 0, 0, 1]
    >>> z_array('ABCA', 'DD')
    [0, 0, 0, 0]

    Args:
        text(str) : text for searching in

        pattern(str): pattern for searching in the text

    Returns all occurences of the pattern in the given text.
    """
    lhs, rhs = [0] * 2
    concatenated = '$'.join([pattern, text])
    z_values = [0] * len(concatenated)
    for i in range(1, len(concatenated)):
        if i > rhs:
            lprefix = longest_prefix(concatenated, pattern, i)
            z_values[i] = len(lprefix)

            if z_values[i] > 0:
                lhs = i
                rhs = i + len(lprefix) - 1
        else:
            k = i - lhs

            if z_values[k] < rhs - i + 1:
                z_values[i] = z_values[k]
            else:
                lhs = i
                lprefix = longest_prefix(concatenated, pattern, i)
                rhs = i + len(lprefix) - 1
                z_values[i] = len(lprefix)

    return z_values[len(pattern) + 1:]



from operator import itemgetter

from pattmatch.helpers import difference

def boyer_moore(text, pattern):
    """
    Finds all start-end positions, containing every start and end of the given pattern in the text.

    >>> boyer_moore('AAA', 'A')
    [[0, 1], [1, 2], [2, 3]]
    >>> boyer_moore('ABAACD', 'AC')
    [[3, 5]]
    >>> boyer_moore('AABAACAADAABAABA', 'AABA')
    [[0, 4], [9, 13], [12, 16]]
    >>> boyer_moore('AAABACAAABCAADCVABCDABCFGABC', 'ABC')
    [[8, 11], [16, 19], [20, 23], [25, 28]]

    Args:
        text(str) : text for searching in
        
        pattern(str): pattern for searching in the text

    Returns all start-end positions of the pattern in the given text.
    """
    start_end_positions = []
    if text == pattern:
        start_end_positions.append([0, len(pattern)])

    i = len(pattern)
    while i < len(text) + 1:
        if text[i-len(pattern):i] == pattern:
            start_end_positions.append([i-len(pattern), i])
        i += difference(text, pattern, i) or 1
    return start_end_positions
